/*  plumb - the pipe shell - common hel;pers for parsing plumb scripts
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "parse_common.h"
#include "plumb.h"
#include "config.h"

char *strclone(const char *s)
{
	char *out;
	int len;

	if (s == NULL)
		return NULL;

	len = strlen(s);
	out = malloc(len+1);
	memcpy(out, s, len+1);
	return out;
}

char *str_aconcat(char *s1, char *s2)
{
	int l1, l2;
	char *out;

	l1 = strlen(s1);
	l2 = strlen(s2);
	out = malloc(l1 + l2 + 1);
	memcpy(out, s1, l1);
	memcpy(out + l1, s2, l2+1);
	free(s1);
	free(s2);
	return out;
}

char *anon_id(void)
{
	static int anon_ctr = 0;
	char buf[128];

	/* buffer is large enough for an int; snprintf() is not portable */
	sprintf(buf, "__plumb_anon_%d__", anon_ctr);
	anon_ctr++;
	return strclone(buf);
}

int strkeyeq(char *a, char *b)
{
	char *pa = (char *)a;
	char *pb = (char *)b;

	for (; *pa == *pb; pa++, pb++)
		if (*pa == '\0')
			return 1;
	return 0;
}


#define in_range(d, low, high) (((d) >= low) && ((d) <= high))

static int hex_error(void)
{
	plumb_error("Invalid hex digit in string\n");
	return -1;
}

#define xdigit(d) \
	(in_range(d, '0', '9') ? (d) - '0' : \
		(in_range(d, 'a', 'f') ? (d) - 'a' + 10 : \
			(in_range(d, 'A', 'F') ? (d) - 'A' + 10 : hex_error()) \
		) \
 )

unsigned char chex(const char *digits)
{
	return (xdigit(digits[0]) << 4) + xdigit(digits[1]);
}

char *str_quote(const char *si, char qchar)
{
	char *so, *ret;
	int len;

	len = strlen(si);
	ret = so = malloc(len);
	len -= 2;
	for(si++; len > 0; len--,si++) {
		if (*si == '\\') {
			si++;
			switch(*si) {
				case 'n': *so = '\n'; break;
				case 'r': *so = '\r'; break;
				case 't': *so = '\t'; break;
				case 'x': *so = (char)chex(si+1); si += 2; len -= 2; break;
				default: /* for at least ", ' and \ */
					*so = *si;
			}
			len--; 
		}
		else
			*so = *si;
		so++;
	}
	if ((so > ret) && (so[-1] == qchar))
		so[-1] = '\0';
	*so = '\0';
	return ret;
}

char *template_prefix(const char *prefix, char *id)
{
	char *s;
	int pl, il;
	if (prefix == NULL)
		return id;

	if (*id == '.') {
		memmove(id, id+1, strlen(id));
		return id;
	}


	pl = strlen(prefix);
	il = strlen(id);
	s = malloc(pl + il + 2);
	memcpy(s, prefix, pl);
	s[pl] = '.';
	memcpy(s+pl+1, id, il+1);
	free(id);
	return s;
}

void plumb_parse_file(transaction_t *tx, const char *templ, const char *inp_name, FILE *in, fd_t *query_out)
{
	parser_t parser;

	parser.templ = templ;
	parser.inp_name = inp_name;
	parser.tx = tx_this_or_new(tx, inp_name, query_out);
	yylex_init(&parser.scanner);
	yyset_in(in, parser.scanner);
	yyrestart(in, parser.scanner);
	yyparse(&parser);
	tx_finish(parser.tx);
	yylex_destroy(parser.scanner);
}

parser_t *plumb_parse_stream_open(transaction_t *tx, const char *templ, const char *inp_name, fd_t *query_out)
{
	parser_t *parser;

	parser = calloc(1, sizeof(parser_t));
	parser->templ = templ;
	parser->inp_name = inp_name;
	parser->tx = tx_this_or_new(tx, inp_name, query_out);
	parser->state = yypstate_new();
	yylex_init(&parser->scanner);
	return parser;
}

void plumb_parse_stream_buffer(parser_t *parser, const char *data, int len)
{
	int prev, current;
	int status, l, leftover_offs;
	char buff[MAX_TOKEN_LEN];
	const char *start, *end;
/*	char *prevptr;*/
	YYSTYPE lval_prev;


	start = data;
	end = data + len;
	/* split up large input buffer into a smaller temporary buffer and insert
	   leftover before the next portion. We copy data only once. */
	while(start < end) {
		l = sizeof(buff)-2;
		if (parser->stream_leftover != NULL) {
			leftover_offs = strlen(parser->stream_leftover);
			if (leftover_offs >= l)
				plumb_error("stream buffer (len=%d) too small or token too large to fit in\n", sizeof(buff));
			memcpy(buff, parser->stream_leftover, leftover_offs);
			free(parser->stream_leftover);
			parser->stream_leftover = NULL;
			l -= leftover_offs;
		}
		else
			leftover_offs = 0;

		if (start + l > end)
			l = end - start;
		memcpy(buff + leftover_offs, start, l);
		buff[l+leftover_offs] = '\0';
		buff[l+leftover_offs+1] = '\0';
/*		fprintf(stderr, "buff='%s'\n", buff);*/
		start += l;
/*		fprintf(stderr, "S '%s'/%d l=%d leftover=%d start='%s'/%d\n", buff, l+leftover_offs+2, l, leftover_offs, start, end-start);*/
		yy_scan_buffer(buff, l+leftover_offs+2, parser->scanner);
/*		prevptr = NULL; */
		prev = -1;
		/* tokenize the buffer and feed the push-parser with 1 token lookahead:
		   always push the previous token after looking at current token; if current
		   token is end-of-buffer and previous token is a multichar token, it is
		   likely a partial token of a short read. No way to find it unless we
		   read some more input. Store the partial token in the leftover buffer
		   and exit; next iteration insert the leftover in front of the buffer */
		do {
			YYSTYPE lval;
			YYLTYPE lloc;
			current = yylex(&lval, &lloc, parser->scanner, parser);
/*			fprintf(stderr, "current=%d prev=%d '%s' yytext='%s' prevptr='%s' delta=%d\n", current, prev, lval.str, lex_getpos(parser->scanner), prevptr, prevptr - buff);*/

			/* end-of buffer - avoid partial tokens */
			if (current == 0) {
				/* previous token may be partial unless it is a single character long token */
				if ((prev != -1) && (lex_is_token_singlechar(prev))) {
/*					fprintf(stderr, " => %d/'%s'\n", prev, lval_prev.str);*/
					status = yypush_parse(parser->state, prev, &lval_prev, &lloc, parser);
					parser->stream_leftover = NULL;
				}
				else
					parser->stream_leftover = lval.str;
				yypop_buffer_state (parser->scanner); /* close current buffer */
				break;
			}

			/* push previous token, if any */
			if (prev != -1) {
/*				fprintf(stderr, " -> %d/'%s'\n", prev, lval_prev.str);*/
				status = yypush_parse(parser->state, prev, &lval_prev, &lloc, parser);
			}
			else
				status = YYPUSH_MORE; /* at first token - read more */
			prev = current;
			/*prevptr =*/ lex_getpos(parser->scanner);
			lval_prev = lval;
		} while(status == YYPUSH_MORE);
/*		fprintf(stderr, "E '%s'\n", parser->stream_leftover);*/
	}
/*	fprintf(stderr, "------\n"); */
}

void plumb_parse_stream_close(parser_t *parser)
{
	tx_discard(parser->tx);
	yylex_destroy(parser->scanner);
	yypstate_delete(parser->state);
	if (parser->inp_name != NULL)
		free((char *)parser->inp_name);
	free(parser);
}

void plumb_parse_string(transaction_t *tx, const char *templ, const char *inp_name, const char *str, fd_t *query_out)
{
	parser_t parser;

	parser.templ = templ;
	parser.inp_name = inp_name;
	parser.tx = tx_this_or_new(tx, inp_name, query_out);

	yylex_init(&parser.scanner);
	yy_scan_string(str, parser.scanner);
	yyparse(&parser);
	tx_finish(parser.tx);
	yylex_destroy(parser.scanner);
}

void yy_include(transaction_t *tx, const char *templ, char *fn)
{
	FILE *f;
	char *path;

	path = plumb_lib_search(fn);
	if (path == NULL)
		plumb_error("Can't find file %s (included) on library search path\n", fn);
	f = fopen(path, "r");
	if (f == NULL)
		plumb_error("Can't open file %s (included)\n", fn);
	plumb_parse_file(tx, templ, path, f, NULL);
	fclose(f);
	free(path);
}
